"""ocp_vscode - OCC viewer for VSCode"""

#
# Copyright 2023 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__version__ = "2.6.1"

import os

from .show import *
from .config import *
from .comms import *

from .colors import *
from .animation import Animation
from ocp_tessellate.cad_objects import ImageFace

try:
    from ocp_tessellate.tessellator import (
        enable_native_tessellator,
        disable_native_tessellator,
        is_native_tessellator_enabled,
    )

    if os.environ.get("NATIVE_TESSELLATOR") == "0":
        disable_native_tessellator()
    else:
        enable_native_tessellator()

    print(
        "Found and enabled native tessellator.\n"
        "To disable, call `disable_native_tessellator()`\n"
        "To enable, call `enable_native_tessellator()`\n"
    )
except:
    pass



import orjson
import socket
import yaml
from pathlib import Path
from flask import Flask, render_template
from flask_sock import Sock
from simple_websocket import ConnectionClosed
from ocp_vscode.comms import MessageType
from ocp_vscode.backend import ViewerBackend
from ocp_vscode.backend_logo import logo
from ocp_vscode.state import resolve_path

CONFIG_FILE = "~/.ocpvscode_standalone"

DEFAULTS = {
    "debug": False,
    "no_glass": False,
    "no_tools": False,
    "tree_width": 240,
    "theme": "light",
    "control": "trackball",
    "modifier_keys": {
        "shift": "shiftKey",
        "ctrl": "ctrlKey",
        "meta": "metaKey",
    },
    "new_tree_behavior": True,
    "pan_speed": 0.5,
    "rotate_speed": 1.0,
    "zoom_speed": 0.5,
    "axes": False,
    "axes0": False,
    "grid_xy": False,
    "grid_yz": False,
    "grid_xz": False,
    "perspective": False,
    "transparent": False,
    "black_edges": False,
    "collapse": "R",
    "up": "Z",
    "ticks": 10,
    "center_grid": False,
    "default_opacity": 0.5,
    "explode": False,
    "default_edgecolor": "#808080",
    "default_color": "#e8b024",
    "default_thickedgecolor": "MediumOrchid",
    "default_facecolor": "Violet",
    "default_vertexcolor": "MediumOrchid",
    "angular_tolerance": 0.2,
    "deviation": 0.1,
    "ambient_intensity": 1.0,
    "direct_intensity": 1.1,
    "metalness": 0.3,
    "roughness": 0.65,
}

SCRIPTS = """
    <script type="module" src="static/js/three-cad-viewer.esm.js"></script>
    <script type="module" srv="static/js/comms.js"></script>
    <script type="module" srv="static/js/logo.js"></script>
"""

JS = "./static/js/three-cad-viewer.esm.js"
CSS = "./static/css/three-cad-viewer.css"

STATIC = """
        import { Comms } from "./static/js/comms.js";
        import { logo } from "./static/js/logo.js";
"""


def COMMS(host, port):
    return f"""
        const comms = new Comms("{host}", {port});
        const vscode = {{postMessage: (msg) => {{
                comms.sendStatus(msg);
            }}
        }};
        const standaloneViewer = () => {{
            const ocpLogo = JSON.parse(logo);
            decode(ocpLogo);
            
            viewer = showViewer(ocpLogo.data.shapes, ocpLogo.config);
        }}
        window.showViewer = standaloneViewer;
    """


INIT = """onload="showViewer()" """


class Viewer:
    def __init__(self, params):
        self.status = {}
        self.config = {}

        self.configure(params)

        self.app = Flask(__name__)
        self.sock = Sock(self.app)

        self.backend = ViewerBackend(self.port)

        self.python_client = None
        self.javascript_client = None
        self.splash = True

        self.sock.route("/")(self.handle_message)
        self.app.add_url_rule("/viewer", "viewer", self.index)

        # get ip address of current machine
        if params.get("host") == "127.0.0.1":
            self.ip_address = "127.0.0.1"
        elif params.get("host") == "0.0.0.0":
            hostname = socket.gethostname()
            self.ip_address = socket.gethostbyname(hostname)
        else:
            self.ip_address = params.get("host")

    def debug_print(self, *msg):
        if self.debug:
            print("Debug:", *msg)

    def configure(self, params):
        # Start with defaults
        local_config = DEFAULTS.copy()

        # Then apply everything from the config file if it exists
        config_file = Path(resolve_path(CONFIG_FILE))
        if config_file.exists():
            with open(config_file, "r") as f:
                defaults = yaml.safe_load(f)
                for k, v in defaults.items():
                    local_config[k] = v

        # Get all params != their default value and apply it
        grid = [False, False, False]
        for k, v in params.items():
            if k == "port":
                self.port = v
            elif k == "host":
                self.host = v
            elif k not in ["create_configfile"]:
                if v != DEFAULTS.get(k):
                    if k == "grid_xy":
                        grid[0] = True
                    elif k == "grid_yz":
                        grid[1] = True
                    elif k == "grid_xz":
                        grid[2] = True
                    else:
                        local_config[k] = v
        local_config["grid"] = grid

        for k, v in local_config.items():
            if k in ["grid_xy", "grid_yz", "grid_xz"]:
                continue
            if k == "debug":
                self.config["debug"] = self.debug = v
            elif k == "collapse":
                self.config["collapse"] = str(v)
            elif k == "no_glass":
                self.config["glass"] = not v
            elif k == "no_tools":
                self.config["tools"] = not v
            elif k == "perspective":
                self.config["ortho"] = not v

            else:
                self.config[k] = v

        self.debug_print("Config:", self.config)

    def start(self):
        self.app.run(debug=self.debug, port=self.port, host=self.host)
        self.sock.init_app(self.app)
        self.backend.load_model(logo)

    def index(self):
        return render_template(
            "viewer.html",
            standalone_scripts=SCRIPTS,
            standalone_imports=STATIC,
            standalone_comms=COMMS(self.ip_address, self.port),
            standalone_init=INIT,
            styleSrc=CSS,
            scriptSrc=JS,
            treeWidth=self.config["tree_width"],
            **self.config,
        )

    def handle_message(self, ws):

        while True:
            data = ws.receive()
            if isinstance(data, bytes):
                data = data.decode("utf-8")

            message_type = data[0]
            data = data[2:]

            if message_type == "C":
                self.python_client = ws
                cmd = orjson.loads(data)
                if cmd == "status":
                    self.debug_print("Received status command")
                    self.python_client.send(orjson.dumps({"text": self.status}))
                elif cmd == "config":
                    self.debug_print("Received config command")
                    self.config["_splash"] = self.splash
                    self.python_client.send(orjson.dumps(self.config))
                elif cmd.type == "screenshot":
                    self.debug_print("Received screenshot command")
                    self.python_client(orjson.dumps(cmd))

            elif message_type == "D":
                self.python_client = ws
                self.debug_print("Received a new model")
                self.javascript_client.send(data)
                if self.splash:
                    self.splash = False

            elif message_type == "U":
                self.javascript_client = ws
                changes = orjson.loads(data)["text"]
                self.debug_print("Received incremental UI changes", changes)
                for key, value in changes.items():
                    self.status[key] = value
                self.backend.handle_event(changes, MessageType.UPDATES)

            elif message_type == "S":
                self.python_client = ws
                self.debug_print("Received a config")
                self.javascript_client.send(data)
                self.debug_print("Posted config to view")

            elif message_type == "L":
                self.javascript_client = ws
                self.debug_print("Javascript listener registered", data)

            elif message_type == "B":
                model = orjson.loads(data)["model"]
                self.backend.handle_event(model, MessageType.DATA)
                self.debug_print("Model data sent to the backend")

            elif message_type == "R":
                self.python_client = ws
                self.javascript_client.send(data)
                self.debug_print("Backend response received.", data)



from build123d import *
from ocp_vscode import *

# set_port(3939)
# set_defaults(reset_camera=Camera.KEEP, ortho=True)

with BuildPart() as p:
    Box(10, 10, 10)
    with Locations((5, 0, 0)):
        Box(10, 10, 5)
c = Circle(10)
cc = extrude(c, 10)
for e in cc.faces()[0].edges():
    print(e.geom_type())
v = Vertex(0, 0, 0)
# print(c.geom_type())
pass
# line = line.edges()[0]
# show(p, f1, f2)



import click
import socket
import yaml
from pathlib import Path
from ocp_vscode.standalone import Viewer, DEFAULTS, CONFIG_FILE
from ocp_vscode.state import resolve_path


def represent_list(dumper, data):
    return dumper.represent_sequence("tag:yaml.org,2002:seq", data, flow_style=True)


yaml.add_representer(list, represent_list)


@click.command()
@click.option(
    "--create_configfile",
    is_flag=True,
    help="Create the configlie .ocpvscode_standalone in the home directory",
)
@click.option(
    "--host",
    default="127.0.0.1",
    help="The host to start OCP CAD with",
)
@click.option(
    "--port",
    default=3939,
    help="The port to start OCP CAD with",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Show debugging information",
)
@click.option(
    "--timeit",
    is_flag=True,
    help="Show timing information",
)
@click.option(
    "--tree_width",
    help="OCP CAD Viewer navigation tree width (default: 240)",
)
@click.option(
    "--no_glass",
    is_flag=True,
    help="Do not use glass mode with transparent navigation tree",
)
@click.option(
    "--theme",
    default="light",
    help="Use theme 'light' or 'dark' (default: 'light')",
)
@click.option(
    "--no_tools",
    is_flag=True,
    help="Do not show toolbar",
)
@click.option(
    "--tree_width", default=240, help="Width of the CAD navigation tree (default: 240)"
)
@click.option(
    "--control",
    default="trackball",
    help="Use control mode 'orbit'or 'trackball'",
)
@click.option(
    "--up",
    default="Z",
    help="Provides up direction, 'Z', 'Y' or 'L' (legacy) (default: Z)",
)
@click.option(
    "--rotate_speed",
    default=1,
    help="Rotation speed (default: 1)",
)
@click.option(
    "--zoom_speed",
    default=1,
    help="Zoom speed (default: 1)",
)
@click.option(
    "--pan_speed",
    default=1,
    help="Pan speed (default: 1)",
)
@click.option(
    "--axes",
    is_flag=True,
    help="Show axes",
)
@click.option(
    "--axes0",
    is_flag=True,
    help="Show axes at the origin (0, 0, 0)",
)
@click.option(
    "--black_edges",
    is_flag=True,
    help="Show edges in black",
)
@click.option(
    "--grid_xy",
    is_flag=True,
    help="Show grid on XY plane",
)
@click.option(
    "--grid_yz",
    is_flag=True,
    help="Show grid on YZ plane",
)
@click.option(
    "--grid_xz",
    is_flag=True,
    help="Show grid on XZ plane",
)
@click.option(
    "--center_grid",
    is_flag=True,
    help="Show grid planes crossing at center of object or global origin(default: False)",
)
@click.option(
    "--collapse",
    default=1,
    help="leaves: collapse all leaf nodes, all: collapse all nodes, none: expand all nodes, root: expand root only (default: leaves)",
)
@click.option(
    "--perspective",
    is_flag=True,
    help="Use perspective camera",
)
@click.option(
    "--ticks",
    default=10,
    help="Default number of ticks (default: 10)",
)
@click.option(
    "--transparent",
    is_flag=True,
    help="Show objects transparent",
)
@click.option(
    "--default_opacity",
    default=0.5,
    help="Default opacity for transparent objects (default: 0.5)",
)
@click.option(
    "--explode",
    is_flag=True,
    help="Turn explode mode on",
)
@click.option(
    "--angular_tolerance",
    default=0.2,
    help="Angular tolerance for tessellation algorithm (default: 0.2)",
)
@click.option(
    "--deviation",
    default=0.1,
    help="Deviation of for tessellation algorithm (default: 0.1)",
)
@click.option(
    "--default_color",
    default="#e8b024",
    help="Default shape color, CSS3 color names are allowed (default: #e8b024)",
)
@click.option(
    "--default_edgecolor",
    default="#707070",
    help="Default color of the edges of shapes, CSS3 color names are allowed (default: #707070)",
)
@click.option(
    "--default_thickedgecolor",
    default="MediumOrchid",
    help="Default color of lines, CSS3 color names are allowed (default: MediumOrchid)",
)
@click.option(
    "--default_facecolor",
    default="Violet",
    help="Default color of faces, CSS3 color names are allowed (default: Violet)",
)
@click.option(
    "--default_vertexcolor",
    default="MediumOrchid",
    help="Default color of vertices, CSS3 color names are allowed (default: MediumOrchid)",
)
@click.option(
    "--ambient_intensity",
    default=1,
    help="Intensity of ambient light (default: 1.00)",
)
@click.option(
    "--direct_intensity",
    default=1.1,
    help="Intensity of direct light (default: 1.10)",
)
@click.option(
    "--metalness",
    default=0.3,
    help="Metalness property of material (default: 0.30)",
)
@click.option(
    "--roughness",
    default=0.65,
    help="Roughness property of material (default: 0.65)",
)
def main(*args, **kwargs):
    if kwargs.get("create_configfile"):

        config_file = Path(resolve_path(CONFIG_FILE))
        with open(config_file, "w") as f:
            f.write(yaml.dump(DEFAULTS))
        print(f"Created config file {config_file}")

    else:
        viewer = Viewer(kwargs)

        port = kwargs["port"]
        host = kwargs["host"]

        if host == "0.0.0.0":
            hostname = socket.gethostname()
            host = socket.gethostbyname(hostname)

        print(f"\nThe viewer is running at http://{host}:{port}/viewer\n")

        viewer.start()


if __name__ == "__main__":
    main()



"""Communication with the viewer"""

import base64
import enum
import json
import os
import socket
import warnings

from pathlib import Path

from websockets.sync.client import connect

import orjson
from ocp_tessellate.utils import Timer
from ocp_tessellate.ocp_utils import (
    is_topods_shape,
    is_toploc_location,
    serialize,
    loc_to_tq,
)
from .state import get_state, update_state, get_config_file


# pylint: disable=unused-import
try:
    from IPython import get_ipython
    import jupyter_console

    JCONSOLE = True
except:  # pylint: disable=bare-except
    JCONSOLE = False

CMD_URL = "ws://127.0.0.1"
CMD_PORT = 3939

INIT_DONE = False


warnings.simplefilter("once", UserWarning)


def warn_once(message):
    def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
        return "%s: %s\n" % (category.__name__, message)

    warnings.formatwarning = warning_on_one_line
    warnings.warn(message, UserWarning)


#
# Send data to the viewer
#


class MessageType(enum.IntEnum):
    """Message types"""

    DATA = 1
    COMMAND = 2
    UPDATES = 3
    LISTEN = 4
    BACKEND = 5
    BACKEND_RESPONSE = 6
    CONFIG = 7


__all__ = [
    "send_data",
    "send_command",
    "send_response",
    "set_port",
    "get_port",
    "listener",
    "is_pytest",
]


def is_pytest():
    return os.environ.get("OCP_VSCODE_PYTEST") == "1"


def port_check(port):
    """Check whether the port is listening"""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(1)
    result = s.connect_ex((get_host(), port)) == 0
    if result:
        s.close()
    return result


def default(obj):
    """Default JSON serializer."""
    if is_topods_shape(obj):
        return base64.b64encode(serialize(obj)).decode("utf-8")
    elif is_toploc_location(obj):
        return loc_to_tq(obj)
    elif isinstance(obj, enum.Enum):
        return obj.value
    else:
        raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def get_port():
    """Get the port"""
    if is_pytest():
        return 3939

    if not INIT_DONE:
        find_and_set_port()
        set_connection_file()
    return CMD_PORT


def get_host():
    """Get the host"""
    return CMD_URL[5:]


def set_port(port, host="127.0.0.1"):
    """Set the port"""
    global CMD_PORT, CMD_URL, INIT_DONE  # pylint: disable=global-statement
    CMD_PORT = port
    CMD_URL = f"ws://{host}"
    INIT_DONE = True


def _send(data, message_type, port=None, timeit=False):
    """Send data to the viewer"""
    global WS

    if port is None:
        if not INIT_DONE:
            find_and_set_port()
            set_connection_file()
        port = CMD_PORT
    try:
        with Timer(timeit, "", "json dumps", 1):
            j = orjson.dumps(data, default=default)  # pylint: disable=no-member
            if message_type == MessageType.COMMAND:
                j = b"C:" + j
            elif message_type == MessageType.DATA:
                j = b"D:" + j
            elif message_type == MessageType.LISTEN:
                j = b"L:" + j
            elif message_type == MessageType.BACKEND:
                j = b"B:" + j
            elif message_type == MessageType.BACKEND_RESPONSE:
                j = b"R:" + j
            elif message_type == MessageType.CONFIG:
                j = b"S:" + j

        with Timer(timeit, "", f"websocket connect ({message_type.name})", 1):
            try:
                with connect(f"{CMD_URL}:{port}", close_timeout=0.05) as ws:
                    ws.send(j)

                    with Timer(
                        timeit, "", f"websocket send {len(j)/1024/1024:.3f} MB", 1
                    ):
                        result = None
                        if message_type == MessageType.COMMAND and not (
                            isinstance(data, dict) and data["type"] == "screenshot"
                        ):
                            try:
                                result = json.loads(ws.recv())
                            except Exception as ex:  # pylint: disable=broad-except
                                print(ex)

            except Exception as ex:
                warn_once("The viewer doesn't seem to run: " + str(ex))
                # set some dummy values to avoid errors
                return {
                    "collapse": "none",
                    "_splash": False,
                    "default_facecolor": (1, 234, 56),
                    "default_thickedgecolor": (123, 45, 6),
                    "default_vertexcolor": (123, 45, 6),
                }

        return result

    except Exception as ex:  # pylint: disable=broad-except
        print(
            f"Cannot connect to viewer on port {port}, is it running and the right port provided?"
        )
        print(ex)
        return None


def send_data(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.DATA, port, timeit)


def send_config(config, port=None, timeit=False):
    """Send config to the viewer"""
    return _send(config, MessageType.CONFIG, port, timeit)


def send_command(data, port=None, timeit=False):
    """Send command to the viewer"""
    return _send(data, MessageType.COMMAND, port, timeit)


def send_backend(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.BACKEND, port, timeit)


def send_response(data, port=None, timeit=False):
    """Send data to the viewer"""
    return _send(data, MessageType.BACKEND_RESPONSE, port, timeit)


#
# Receive data from the viewer
#


# async listerner for the websocket class
# this will be called when the viewer sends data
# the data is then passed to the callback function
#
def listener(callback):
    """Listen for data from the viewer"""

    def _listen():
        last_config = {}
        with connect(f"{CMD_URL}:{CMD_PORT}", max_size=2**28) as websocket:
            websocket.send(b"L:register")
            while True:
                try:
                    message = websocket.recv()
                    if message is None:
                        continue

                    message = json.loads(message)
                    if "model" in message.keys():
                        callback(message["model"], MessageType.DATA)

                    if message.get("command") == "status":
                        changes = message["text"]
                        new_changes = {}
                        for k, v in changes.items():
                            if k in last_config and last_config[k] == v:
                                continue
                            new_changes[k] = v
                        last_config = changes
                        callback(new_changes, MessageType.UPDATES)

                except Exception as ex:  # pylint: disable=broad-except
                    print(ex)
                    break

    return _listen


def find_and_set_port():
    """Set the port and connection file"""

    def find_port():
        port = None
        current_path = Path.cwd()
        states = get_state().items()
        for p, state in states:
            if not port_check(int(p)):
                print(f"Found stale configuration for port {p}, deleting it.")
                update_state(int(p), None, None)
                continue

            roots = state.get("roots", [])
            for root in roots:
                if current_path.is_relative_to(Path(root)):
                    port = int(p)
                    break
        if port is None:
            ports = [port for port, _ in states if port_check(int(port))]
            if len(ports) == 1:
                port = ports[0]
            elif len(ports) > 1:
                raise RuntimeError(
                    f"\nMultiple ports found ({', '.join(ports)}) and the file is outside of any\n"
                    "workspace folder of this VS Code instance:\n"
                    "The right viewer cannot be auto detected, use set_port(port) in your code."
                )
            else:
                warn_once("The viewer doesn't seem to run")
                # raise RuntimeError(
                #     "No port found via config file\n"
                #     "To change the port, use set_port(port) in your code"
                # )

        return port

    port = int(os.environ.get("OCP_PORT", "0"))

    if port > 0:
        print(f"Using predefined port {port} taken from environment variable OCP_PORT")
    else:
        port = find_port()
        if port is not None:
            print(f"Using port {port} taken from config file")
        elif port_check(3939):
            port = 3939
            print(f"Port {port} is open, using it")

    set_port(port)


def set_connection_file():
    """Set the connection file for Jupyter in the state file .ocpvscode"""
    if JCONSOLE and hasattr(get_ipython(), "kernel"):
        kernel = get_ipython().kernel
        cf = kernel.config["IPKernelApp"]["connection_file"]
        with open(cf, "r", encoding="utf-8") as f:
            connection_info = json.load(f)

        if port_check(connection_info["iopub_port"]):
            print("Jupyter kernel running")
            update_state(CMD_PORT, "connection_file", cf)
            print(f"Jupyter connection file path written to {get_config_file()}")
        else:
            print("Jupyter kernel not responding")
    elif not JCONSOLE:
        print("Jupyter console not installed")
    else:
        print("Jupyter kernel not running")



from build123d import *
from ocp_vscode import *

set_port(3939)
set_defaults(reset_camera=Camera.KEEP, ortho=True)
densa = 7800 / 1e6  # carbon steel density g/mm^3
densb = 2700 / 1e6  # aluminum alloy
densc = 1020 / 1e6  # ABS
LB = 453.592  # g/lb
ms = Mode.SUBTRACT
LMH, LMV = LengthMode.HORIZONTAL, LengthMode.VERTICAL
# %%

with BuildPart() as p:
    with BuildSketch(Plane.XZ) as s:
        with BuildLine(Plane.XZ) as l:
            m1 = Line((3.5 / 2, 0), (5.625 / 2, 0))
            m2 = Line(m1 @ 1, m1 @ 1 + (0, 3))
            m3 = Line(m2 @ 1, m2 @ 1 + (-(5.625 - 3.75) / 2, 0))
            m4 = Line(m3 @ 1, m3 @ 1 + (0, -0.375))
            m5 = Line(m4 @ 1, m4 @ 1 + (-0.75 / 2, 0))
            m6 = Line(m5 @ 1, m5 @ 1 + (0, -3 + 1.75 + 0.375))
            m7 = Line(m6 @ 1, m6 @ 1 + (0.5 / 2, 0))
            m8 = Line(m7 @ 1, m1 @ 0)
        make_face()
    revolve(axis=Axis.Z, revolution_arc=90)

    with BuildSketch(Plane.XZ.rotated((0, 0, -45))) as s2:
        Rectangle(1, 1.5, align=(Align.CENTER, Align.MIN))
        with Locations((0, 1)):
            Rectangle(2, 0.25, align=(Align.CENTER, Align.MIN))
    extrude(amount=-6, mode=ms)

    with Locations(Plane.XY.offset(3)):
        with PolarLocations(4.625 / 2, 1, 45):
            CounterBoreHole(0.375 / 2, 0.625 / 2, 0.25)
    mirror(about=Plane.XZ)
    mirror(about=Plane.YZ)


r = RegularPolygon(100, 5)
pp = extrude(r, 1000)
pp -= Hole(2, 50)


classes = (BuildPart, BuildSketch, BuildLine)  # for OCP-vscode
set_colormap(ColorMap.seeded(colormap="rgb", alpha=1, seed_value="vscod"))
show2(pp, name="item")

# b = Box(1, 2, 3) - Plane.YZ * Cylinder(0.5, 1)
# b = fillet(b.edges().filter_by(Axis.X), 0.3)
# b = chamfer(b.edges().filter_by(Axis.Y), 0.1)
# c = Pos(3, 0, 0) * Box(1, 1, 1)
# s = Pos(0, 2, 1) * Circle(0.5)

# show2(b, c, s)
print()
print(f"part mass = {p.part.scale(IN).volume*densb/LB}")



"""Python backend for the OCP Viewer"""

import argparse
import base64
import sys
import traceback
from dataclasses import asdict, dataclass, fields

from ocp_tessellate.ocp_utils import (
    deserialize,
    make_compound,
    tq_to_loc,
    get_surface,
    get_curve,
)
from ocp_tessellate.tessellator import (
    get_edges,
    get_faces,
    get_vertices,
)
from ocp_tessellate.trace import Trace

from ocp_vscode.backend_logo import logo
from ocp_vscode.build123d import (
    Compound,
    Edge,
    Face,
    Location,
    Plane,
    Shape,
    Solid,
    Vector,
    Vertex,
    downcast,
)

from ocp_vscode.comms import MessageType, listener, send_response, set_port


class SelectedCenterInfo:
    """
    Stores the information message about what as been used as center for the measurement
    """

    vertex = "Reference point has been taken as vertex location"
    circular = "Reference point has been taken as the center of the circle or ellipse"
    geom = "Reference point has been taken as the center of the geometry"
    cylinder = "Reference point has been taken as the center of the cylinder"


def print_to_stdout(*msg):
    """
    Write the given message to the stdout
    """
    print(*msg, flush=True, file=sys.stdout)


def error_handler(func):
    """Decorator for error handling"""

    def wrapper(*args, **kwargs):  # pylint: disable=redefined-outer-name
        try:
            func(*args, **kwargs)
        except Exception as exc:  # pylint: disable=broad-except
            print_to_stdout(exc)
            traceback.print_exception(*sys.exc_info(), file=sys.stdout)

    return wrapper


@dataclass
class Tool:
    """The tools available in the viewer"""

    Distance = "DistanceMeasurement"
    Properties = "PropertiesMeasurement"
    Angle = "AngleMeasurement"


def set_precision(instance, decimals=2):
    """
    Set the precision of all float fields of the dataclass to the given number of decimals
    """
    for field in fields(instance):
        if field.type == float:
            value = getattr(instance, field.name)
            if value is not None:
                setattr(instance, field.name, round(value, decimals))
        elif isinstance(getattr(instance, field.name), tuple):
            # Handle tuple fields
            old_tuple = getattr(instance, field.name)
            new_tuple = tuple(
                round(elem, decimals) if isinstance(elem, float) else elem
                for elem in old_tuple
            )
            setattr(instance, field.name, new_tuple)


@dataclass
class Response:
    """Base class for all responses"""

    type: str = "backend_response"


@dataclass
class MeasureReponse(Response):
    """Base class for all measurement responses"""

    # a string telling the frontend how the reference points used for the
    # measurement were chosen
    center_info: str = ""
    subtype: str = "tool_response"


@dataclass
class DistanceResponse(MeasureReponse):
    """Response class for distance measurement"""

    tool_type: Tool = Tool.Distance
    point1: tuple = None
    point2: tuple = None
    distance: float = None


@dataclass
class PropertiesResponse(MeasureReponse):
    """Response class for properties measurement"""

    tool_type: Tool = Tool.Properties
    center: tuple = None
    vertex_coords: tuple = None
    length: float = None
    width: float = None
    area: float = None
    volume: float = None
    radius: float = None
    radius2: float = None
    geom_type: str = None


@dataclass
class AngleResponse(MeasureReponse):
    """Response class for angle measurement"""

    tool_type: Tool = Tool.Angle
    angle: float = None
    point1: tuple = None
    point2: tuple = None


class ViewerBackend:
    """
    Represents the backend of the viewer, it listens to the websocket and handles the events
    It's job is to send responses to the vscode extension that goes through the three cad
    viewer view.
    The reponses holds all the data needed to display the measurements.
    """

    def __init__(self, port: int) -> None:
        self.port = port
        self.model = None
        self.activated_tool = None
        self.filter_type = "none"  # The current active selection filter
        set_port(port)

    def start(self):
        """
        Start the backend
        """
        print("Viewer backend started")
        backend.load_model(logo)
        print("Logo model loaded")
        listener(self.handle_event)()

    @error_handler
    def handle_event(self, message, event_type: MessageType):
        """
        Handle the event received from the websocket
        Dispatch the event to the appropriate handler
        """
        if event_type == MessageType.DATA:
            self.load_model(message)
        elif event_type == MessageType.UPDATES:
            changes = message

            if "activeTool" in changes:
                active_tool = changes.get("activeTool")

                if active_tool != "None":
                    self.activated_tool = active_tool
                else:
                    self.activated_tool = None

            if self.activated_tool is not None:
                self.handle_activated_tool(changes)

    def handle_activated_tool(self, changes):
        """
        Handle the activated tool, there is a special behavior for each tool
        """
        if not "selectedShapeIDs" in changes:
            return

        selected_objs = changes["selectedShapeIDs"]
        if self.activated_tool == Tool.Distance and len(selected_objs) == 2:
            shape_id1 = changes["selectedShapeIDs"][0]
            shape_id2 = changes["selectedShapeIDs"][1]
            self.handle_distance(shape_id1, shape_id2)

        elif self.activated_tool == Tool.Properties and len(selected_objs) == 1:
            shape_id = changes["selectedShapeIDs"][0]
            self.handle_properties(shape_id)

        elif self.activated_tool == Tool.Angle and len(selected_objs) == 2:
            shape_id1 = changes["selectedShapeIDs"][0]
            shape_id2 = changes["selectedShapeIDs"][1]
            self.handle_angle(shape_id1, shape_id2)

    def load_model(self, raw_model):
        """Read the transfered model from websocket"""

        def walk(model, trace):
            for v in model["parts"]:
                if v.get("parts") is not None:
                    walk(v, trace)
                else:
                    id_ = v["id"]
                    loc = (
                        Location().wrapped if v["loc"] is None else tq_to_loc(*v["loc"])
                    )
                    if isinstance(v["shape"], dict):
                        compound = deserialize(
                            base64.b64decode(v["shape"]["obj"].encode("utf-8"))
                        )
                    else:
                        shape = [
                            deserialize(base64.b64decode(s.encode("utf-8")))
                            for s in v["shape"]
                        ]
                        compound = make_compound(shape) if len(shape) > 1 else shape[0]
                    self.model[id_] = Compound(compound.Moved(loc))
                    faces = get_faces(compound)
                    for i, face in enumerate(faces):
                        trace.face(f"{id_}/faces/faces_{i}", face)

                        self.model[f"{id_}/faces/faces_{i}"] = Face(face.Moved(loc))
                    edges = get_edges(compound)
                    for i, edge in enumerate(edges):
                        trace.edge(f"{id_}/edges/edges_{i}", edge)

                        self.model[f"{id_}/edges/edges_{i}"] = (
                            Edge(edge) if loc is None else Edge(edge.Moved(loc))
                        )
                    vertices = get_vertices(compound)
                    for i, vertex in enumerate(vertices):
                        trace.vertex(f"{id_}/vertices/vertex{i}", vertex)

                        self.model[f"{id_}/vertices/vertices{i}"] = (
                            Vertex(vertex)
                            if loc is None
                            else Vertex(downcast(vertex.Moved(loc)))
                        )

        self.model = {}
        trace = Trace("ocp-vscode-backend.log")
        walk(raw_model, trace)
        trace.close()

    def handle_properties(self, shape_id):
        """
        Request the properties of the object with the given id
        """
        print_to_stdout(f"Identifier received '{shape_id}'")

        shape = self.model[shape_id]

        response = PropertiesResponse()

        if isinstance(shape, Vertex):
            response.vertex_coords = shape.to_tuple()

        elif isinstance(shape, Edge):
            response.radius = None
            response.major_radius = None
            response.minor_radius = None
            if shape.geom_type == "CIRCLE":
                response.radius = shape.radius
            elif shape.geom_type == "ELLIPSE":
                response.radius = get_curve(shape.wrapped).Ellipse().MajorRadius()
                response.radius2 = get_curve(shape.wrapped).Ellipse().MinorRadius()

            response.length = shape.length

        elif isinstance(shape, Face):
            if shape.geom_type == "CYLINDER":
                response.radius = get_surface(shape.wrapped).Cylinder().Radius()

            response.length = shape.length
            response.width = shape.width
            response.area = shape.area

        elif isinstance(shape, (Solid, Compound)):
            response.volume = shape.volume

        response.geom_type = shape.geom_type if not isinstance(shape, Vertex) else None
        center, info = self.get_center(shape, False)
        response.center = center.to_tuple()
        response.vertex_coords = response.center
        response.center_info = f"{shape_id} : {info}"

        set_precision(response)

        send_response(asdict(response), self.port)
        print_to_stdout(f"Data sent {response}")

    def handle_angle(self, id1, id2):
        """
        Request the angle between the two objects that have the given ids
        """
        print_to_stdout(f"Identifiers received '{id1}', '{id2}'")

        shape1: Shape = self.model[id1]
        shape2: Shape = self.model[id2]
        first = (
            Plane(shape1)
            if isinstance(shape1, Face)
            else (
                Plane(shape1 @ 0, z_dir=shape1.normal())
                if isinstance(shape1, Edge)
                and shape1.geom_type in ["CIRCLE", "ELLIPSE"]
                else shape1 % 0
            )
        )
        second = (
            Plane(shape2)
            if isinstance(shape2, Face)
            else (
                Plane(shape2 @ 0, z_dir=shape2.normal())
                if isinstance(shape2, Edge)
                and shape2.geom_type in ["CIRCLE", "ELLIPSE"]
                else shape2 % 0
            )
        )
        if isinstance(first, Plane) and isinstance(second, Plane):
            angle = first.z_dir.get_angle(second.z_dir)
        elif isinstance(first, Vector) and isinstance(second, Vector):
            angle = first.get_angle(second)
        else:
            vector = first if isinstance(first, Vector) else second
            plane = first if isinstance(first, Plane) else second

            angle = 90 - plane.z_dir.get_angle(vector)
        angle = abs(angle)
        point1, info1 = self.get_center(shape1, True)
        point2, info2 = self.get_center(shape2, True)
        center_info = f"{id1} : {info1}\n{id2} : {info2}"
        response = AngleResponse(
            center_info=center_info,
            angle=angle,
            point1=point1.to_tuple(),
            point2=point2.to_tuple(),
        )
        set_precision(response)
        send_response(asdict(response), self.port)
        print_to_stdout(f"Data sent {response}")

    def get_center(
        self, shape: Shape, for_distance=True
    ) -> tuple[Vector, SelectedCenterInfo]:
        """
        Returns the center vector of the given shape
        Center of the shape depends on the type of the shape and the tool used
        For instance, circle edge center will be on the edge for properties tool
        but at the center of the circle for distance tool
        """
        if isinstance(shape, Vertex):
            return shape.center(), SelectedCenterInfo.vertex
        elif isinstance(shape, Edge):
            if shape.geom_type in [
                "CIRCLE",
                "ELLIPSE",
            ]:
                if for_distance:
                    return shape.arc_center, SelectedCenterInfo.circular
                else:
                    return shape.center(), SelectedCenterInfo.geom

        elif isinstance(shape, Face):
            if shape.geom_type in ["CYLINDER"]:
                if not for_distance:
                    return shape.center(), SelectedCenterInfo.geom

                extremity_edges = shape.edges().filter_by("CIRCLE")
                if len(extremity_edges) == 2:
                    return (
                        extremity_edges[0].arc_center
                        - (
                            extremity_edges[0].arc_center
                            - extremity_edges[-1].arc_center
                        )
                        / 2,
                        SelectedCenterInfo.cylinder,
                    )
                else:
                    try:
                        return (
                            extremity_edges[0].arc_center,
                            SelectedCenterInfo.cylinder,
                        )
                    except IndexError:
                        # cylinder might have bspline extremity edges hence the list would be empty
                        # in that case we default to geom center
                        pass

        return shape.center(), SelectedCenterInfo.geom

    def handle_distance(self, id1, id2):
        """
        Request the distance between the two objects that have the given ids
        """
        print_to_stdout(f"Identifiers received '{id1}', '{id2}'")

        shape1: Shape = self.model[id1]
        shape2: Shape = self.model[id2]
        p1, info1 = self.get_center(shape1)
        p2, info2 = self.get_center(shape2)
        center_info = f"{id1} : {info1}\n{id2} : {info2}"
        dist = (p2 - p1).length
        response = DistanceResponse(
            center_info=center_info,
            point1=p1.to_tuple(),
            point2=p2.to_tuple(),
            distance=dist,
        )
        set_precision(response)
        send_response(asdict(response), self.port)
        print_to_stdout(f"Data sent {response}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser("OCP Viewer Backend")
    parser.add_argument(
        "--port", type=int, required=True, help="Port the viewer listens to"
    )
    args = parser.parse_args()
    backend = ViewerBackend(args.port)
    try:
        backend.start()
    except Exception as ex:  # pylint: disable=broad-except
        print(ex)



"""Configuration of the viewer"""

#
# Copyright 2023 Bernhard Walter
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from enum import Enum

from .comms import send_command, send_config, get_port, is_pytest

__all__ = [
    "workspace_config",
    "combined_config",
    "set_viewer_config",
    "set_defaults",
    "reset_defaults",
    "get_default",
    "get_defaults",
    "status",
    "Camera",
    "Collapse",
    "check_deprecated",
]


class Camera(Enum):
    """Camera reset modes"""

    RESET = "reset"
    CENTER = "center"
    KEEP = "keep"


class Collapse(Enum):
    """Collapse modes for the CAD navigation tree"""

    NONE = 0
    LEAVES = 1
    ALL = 2
    ROOT = 3


COLLAPSE_MAPPING = ["E", "1", "C", "R"]

CONFIG_UI_KEYS = [
    "axes",
    "axes0",
    "black_edges",
    "grid",
    "ortho",
    "transparent",
    "explode",
    "ambient_intensity",
    "direct_intensity",
    "metalness",
    "roughness",
    "clip_slider_0",
    "clip_slider_1",
    "clip_slider_2",
    "clip_normal_0",
    "clip_normal_1",
    "clip_normal_2",
    "clip_planes",
    "clip_intersection",
    "clip_object_colors",
]

CONFIG_WORKSPACE_KEYS = CONFIG_UI_KEYS + [
    # viewer
    "collapse",
    "dark",
    "glass",
    "orbit_control",
    "ticks",
    "center_grid",
    "tools",
    "tree_width",
    "up",
    # mouse
    "pan_speed",
    "rotate_speed",
    "zoom_speed",
    # render settings
    "ambient_intensity",
    "direct_intensity",
    "metalness",
    "roughness",
    "angular_tolerance",
    "default_color",
    "default_edgecolor",
    "default_facecolor",
    "default_thickedgecolor",
    "default_vertexcolor",
    "default_opacity",
    "deviation",
    "modifier_keys",
]

CONFIG_CONTROL_KEYS = [
    "edge_accuracy",
    "debug",
    "helper_scale",
    "render_edges",
    "render_mates",
    "render_joints",
    "render_normals",
    "reset_camera",
    "show_parent",
    "show_sketch_local",
    "timeit",
]

CONFIG_KEYS = CONFIG_WORKSPACE_KEYS + CONFIG_CONTROL_KEYS + ["zoom"]

CONFIG_SET_KEYS = [
    "axes",
    "axes0",
    "grid",
    "ortho",
    "transparent",
    "black_edges",
    "explode",
    "zoom",
    "position",
    "quaternion",
    "target",
    "default_edgecolor",
    "default_opacity",
    "ambient_intensity",
    "direct_intensity",
    "metalness",
    "roughness",
    "zoom_speed",
    "pan_speed",
    "rotate_speed",
    "glass",
    "tools",
    "tree_width",
    "collapse",
    # "tab",
    # "clip_slider_0",
    # "clip_slider_1",
    # "clip_slider_2",
    # "clip_normal_0",
    # "clip_normal_1",
    # "clip_normal_2",
    # "clip_intersection",
    # "clip_planes",
    # "clip_object_colors",
]

DEFAULTS = {
    "render_edges": True,
    "render_normals": False,
    "render_mates": False,
    "render_joints": False,
    "helper_scale": 1.0,
    "show_parent": False,
    "show_sketch_local": True,
    "timeit": False,
    "reset_camera": Camera.RESET,
    "collapse": Collapse.ROOT,
    "debug": False,
}


# pylint: disable=too-many-arguments,unused-argument,too-many-locals
def set_viewer_config(
    axes=None,
    axes0=None,
    grid=None,
    ortho=None,
    transparent=None,
    black_edges=None,
    explode=None,
    zoom=None,
    position=None,
    quaternion=None,
    target=None,
    default_edgecolor=None,
    default_opacity=None,
    ambient_intensity=None,
    direct_intensity=None,
    metalness=None,
    roughness=None,
    zoom_speed=None,
    pan_speed=None,
    rotate_speed=None,
    glass=None,
    tools=None,
    tree_width=None,
    collapse=None,
    reset_camera=None,
    states=None,
    tab=None,
    clip_slider_0=None,
    clip_slider_1=None,
    clip_slider_2=None,
    clip_normal_0=None,
    clip_normal_1=None,
    clip_normal_2=None,
    clip_intersection=None,
    clip_planes=None,
    clip_object_colors=None,
    port=None,
):
    """Set viewer config"""
    if port is None:
        port = get_port()

    config = {k: v for k, v in locals().items() if v is not None}
    if config.get("collapse") is not None:
        config["collapse"] = COLLAPSE_MAPPING[config["collapse"].value]
    data = {
        "type": "ui",
        "config": config,
    }

    try:
        send_config(data, port)

    except Exception as ex:
        raise RuntimeError(
            "Cannot set viewer config. Is the viewer running?\n" + str(ex.args)
        ) from ex


def get_default(key):
    """Get default value for key"""
    return DEFAULTS.get(key)


def get_defaults():
    """Get all defaults"""
    return DEFAULTS


def set_defaults(
    glass=None,
    tools=None,
    tree_width=None,
    axes=None,
    axes0=None,
    grid=None,
    ortho=None,
    transparent=None,
    default_opacity=None,
    black_edges=None,
    orbit_control=None,
    collapse=None,
    ticks=None,
    center_grid=None,
    up=None,
    explode=None,
    zoom=None,
    reset_camera=None,
    clip_slider_0=None,
    clip_slider_1=None,
    clip_slider_2=None,
    clip_normal_0=None,
    clip_normal_1=None,
    clip_normal_2=None,
    clip_intersection=None,
    clip_planes=None,
    clip_object_colors=None,
    pan_speed=None,
    rotate_speed=None,
    zoom_speed=None,
    deviation=None,
    angular_tolerance=None,
    edge_accuracy=None,
    default_color=None,
    default_edgecolor=None,
    ambient_intensity=None,
    direct_intensity=None,
    metalness=None,
    roughness=None,
    render_edges=None,
    render_normals=None,
    render_mates=None,
    render_joints=None,
    show_parent=None,
    show_sketch_local=None,
    helper_scale=None,
    mate_scale=None,  # DEPRECATED
    debug=None,
    timeit=None,
):
    # pylint: disable=line-too-long
    """Set viewer defaults
    Keywords to configure the viewer:
    - UI
        glass:             Use glass mode where tree is an overlay over the cad object (default=False)
        tools:             Show tools (default=True)
        tree_width:        Width of the object tree (default=240)

    - Viewer
        axes:              Show axes (default=False)
        axes0:             Show axes at (0,0,0) (default=False)
        grid:              Show grid (default=False)
        ortho:             Use orthographic projections (default=True)
        transparent:       Show objects transparent (default=False)
        default_opacity:   Opacity value for transparent objects (default=0.5)
        black_edges:       Show edges in black color (default=False)
        orbit_control:     Mouse control use "orbit" control instead of "trackball" control (default=False)
        collapse:          Collapse.LEAVES: collapse all single leaf nodes,
                           Collapse.ROOT: expand root only,
                           Collapse.ALL: collapse all nodes,
                           Collapse.NONE: expand all nodes
                           (default=Collapse.ROOT)
        ticks:             Hint for the number of ticks in both directions (default=10)
        center_grid:       Center the grid at the origin or center of mass (default=False)
        up:                Use z-axis ('Z') or y-axis ('Y') as up direction for the camera (default="Z")
        explode:           Turn on explode mode (default=False)

        zoom:              Zoom factor of view (default=1.0)
        position:          Camera position
        quaternion:        Camera orientation as quaternion
        target:            Camera look at target
        reset_camera:      Camera.RESET: Reset camera position, rotation, toom and target
                           Camera.CENTER: Keep camera position, rotation, toom, but look at center
                           Camera.KEEP: Keep camera position, rotation, toom, and target
                           (default=Camera.RESET)
        clip_slider_0:     Setting of clipping slider 0 (default=None)
        clip_slider_1:     Setting of clipping slider 1 (default=None)
        clip_slider_2:     Setting of clipping slider 2 (default=None)
        clip_normal_0:     Setting of clipping normal 0 (default=[-1,0,0])
        clip_normal_1:     Setting of clipping normal 1 (default=[0,-1,0])
        clip_normal_2:     Setting of clipping normal 2 (default=[0,0,-1])
        clip_intersection: Use clipping intersection mode (default=[False])
        clip_planes:       Show clipping plane helpers (default=False)
        clip_object_colors: Use object color for clipping caps (default=False)

        pan_speed:         Speed of mouse panning (default=1)
        rotate_speed:      Speed of mouse rotate (default=1)
        zoom_speed:        Speed of mouse zoom (default=1)

    - Renderer
        deviation:         Shapes: Deviation from linear deflection value (default=0.1)
        angular_tolerance: Shapes: Angular deflection in radians for tessellation (default=0.2)
        edge_accuracy:     Edges: Precision of edge discretization (default: mesh quality / 100)

        default_color:     Default mesh color (default=(232, 176, 36))
        default_edgecolor: Default mesh color (default=(128, 128, 128))
        ambient_intensity: Intensity of ambient light (default=1.00)
        direct_intensity:  Intensity of direct light (default=1.10)
        metalness:         Metalness property of the default material (default=0.30)
        roughness:         Roughness property of the default material (default=0.65)

        render_edges:      Render edges  (default=True)
        render_normals:    Render normals (default=False)
        render_mates:      Render mates for MAssemblies (default=False)
        render_joints:     Render mates for MAssemblies (default=False)
        show_parent:       Render parent of faces, edges or vertices as wireframe (default=False)
        show_sketch_local: In build123d show local sketch in addition to relocate sketch (default=True)
        helper_scale:      Scale of rendered helpers (locations, axis, mates for MAssemblies) (default=1)

    - Debug
        debug:             Show debug statements to the VS Code browser console (default=False)
        timeit:            Show timing information from level 0-3 (default=False)
    """

    kwargs = {k: v for k, v in locals().items() if v is not None}

    kwargs = check_deprecated(kwargs)

    for key, value in kwargs.items():
        if key in CONFIG_KEYS:
            DEFAULTS[key] = value
        else:
            print(f"'{key}' is an unkown config, ignored!")

    set_viewer_config(**{k: v for k, v in kwargs.items() if k in CONFIG_SET_KEYS})


def preset(key, value):
    """Set default value for key"""
    return get_default(key) if value is None else value


def ui_filter(conf):
    """Filter out all non-UI keys from the config dict"""
    return {k: v for k, v in conf.items() if k in CONFIG_UI_KEYS}


def status(port=None, debug=False):
    """Get viewer status"""

    if is_pytest():
        return {}

    if port is None:
        port = get_port()
    try:
        response = send_command("status", port=port)
        if debug:
            return response.get("_debugStarted", False)
        else:
            return dict(sorted(response.get("text", {}).items()))

    except Exception as ex:
        raise RuntimeError(
            "Cannot access viewer status. Is the viewer running?\n" + str(ex.args)
        ) from ex


def workspace_config(port=None):
    """Get viewer workspace config"""

    if is_pytest():
        return {
            "_splash": False,
            "default_facecolor": (1, 234, 56),
            "default_thickedgecolor": (123, 45, 6),
            "default_vertexcolor": (123, 45, 6),
        }

    if port is None:
        port = get_port()
    try:
        conf = send_command("config", port=port)
        mapping = {
            "none": Collapse.NONE,
            "leaves": Collapse.LEAVES,
            "all": Collapse.ALL,
            "root": Collapse.ROOT,
            "E": Collapse.NONE,
            "1": Collapse.LEAVES,
            "C": Collapse.ALL,
            "R": Collapse.ROOT,
        }
        conf["collapse"] = mapping[conf.get("collapse", "R")]
        return dict(sorted(conf.items()))

    except Exception as ex:
        raise RuntimeError(
            "Cannot access viewer config. Is the viewer running?\n" + str(ex.args)
        ) from ex


def combined_config(port=None, use_status=True):
    """Get combined config from workspace and status"""
    if port is None:
        port = get_port()

    try:
        wspace_config = workspace_config(port)
        wspace_status = status(port)

    except Exception as ex:
        raise RuntimeError(
            "Cannot access viewer config. Is the viewer running?\n" + str(ex.args)
        ) from ex

    if use_status and wspace_config["_splash"]:
        del wspace_config["_splash"]
        wspace_config["axes"] = False
        wspace_config["axes0"] = True
        wspace_config["grid"] = [True, False, False]
        wspace_config["ortho"] = False
        wspace_config["transparent"] = False
        wspace_config["black_edges"] = False

    wspace_config.update(DEFAULTS)
    if use_status:
        wspace_config.update(ui_filter(wspace_status))
    return dict(sorted(wspace_config.items()))


def get_changed_config(key=None):
    """Get changed config from workspace and status"""
    wspace_config = workspace_config()
    wspace_config.update(DEFAULTS)
    if key is None:
        return wspace_config
    else:
        return wspace_config.get(key)


def reset_defaults():
    """Reset defaults not given in workspace config"""
    global DEFAULTS  # pylint: disable=global-statement

    config = {
        key: value
        for key, value in workspace_config().items()
        if key in CONFIG_SET_KEYS
    }
    config["reset_camera"] = Camera.RESET

    set_viewer_config(**config)

    if config.get("transparent") is not None:
        set_viewer_config(transparent=config["transparent"])

    DEFAULTS = {
        "render_edges": True,
        "render_normals": False,
        "render_mates": False,
        "render_joints": False,
        "helper_scale": 1.0,
        "timeit": False,
        "reset_camera": Camera.RESET,
        "debug": False,
    }


def check_deprecated(kwargs):
    """Check for deprecated arguments"""
    if kwargs.get("mate_scale") is not None:
        print("\nmate_scale is deprecated, use helper_scale instead\n")
        kwargs["helper_scale"] = kwargs["mate_scale"]
        del kwargs["mate_scale"]

    if kwargs.get("reset_camera") is True:
        print(
            "\n'reset_camera=True' is deprecated, use 'reset_camera=Camera.RESET' instead\n"
        )
        kwargs["reset_camera"] = Camera.RESET

    if kwargs.get("reset_camera") is False:
        print(
            "\n'reset_camera=False' is deprecated, use 'reset_camera=Camera.CENTER' instead\n"
        )
        kwargs["reset_camera"] = Camera.CENTER

    if kwargs.get("collapse") == "C":
        print("\n'collapse=\"C\"' is deprecated, use 'collapse=Collapse.ALL' instead\n")
        kwargs["collapse"] = Collapse.ALL

    if kwargs.get("collapse") == "1" or kwargs.get("collapse") == 1:
        print(
            "\n'collapse=\"1\"' is deprecated, use 'collapse=Collapse.LEAVES' instead\n"
        )
        kwargs["collapse"] = Collapse.LEAVES

    if kwargs.get("collapse") == "R":
        print(
            "\n'collapse=\"R\"' is deprecated, use 'collapse=Collapse.ROOT' instead\n"
        )
        kwargs["collapse"] = Collapse.ROOT

    if kwargs.get("collapse") == "E":
        print(
            "\n'collapse=\"E\"' is deprecated, use 'collapse=Collapse.NONE' instead\n"
        )
        kwargs["collapse"] = Collapse.NONE

    return kwargs



"""A minimum subset of build123d to support backend.py"""

# pylint: disable=missing-class-docstring,missing-function-docstring,import-error,no-name-in-module
# pyright: reportMissingModuleSource=false

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#              THIS FILE IS AN EXTRACT OF http://github.com/gumyr/build123d
#                      ALL CREDITS GO TO ROGER MAITLAND
#  THE CODE IS SHORTEND/STREAMLINED TO ONLY SUPPORT THE CONCEPTS NEEDED BY backend.py
#                      ALL NEW BUGS ARE OF COURSE MINE!
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import copy
import math
from enum import Enum, auto
from typing import Iterable

import OCP.GeomAbs as ga
import OCP.TopAbs as ta
from OCP.BRep import BRep_Tool
from OCP.BRepAdaptor import BRepAdaptor_Curve, BRepAdaptor_Surface
from OCP.BRepBuilderAPI import BRepBuilderAPI_Transform, BRepBuilderAPI_MakeVertex
from OCP.BRepGProp import BRepGProp, BRepGProp_Face
from OCP.BRepLib import BRepLib_FindSurface
from OCP.BRepTools import BRepTools
from OCP.GCPnts import GCPnts_AbscissaPoint
from OCP.Geom import Geom_Plane
from OCP.gp import (
    gp_Dir,
    gp_Pnt,
    gp_Trsf,
    gp_Vec,
    gp_GTrsf,
    gp_Ax1,
    gp_Ax3,
    gp_EulerSequence,
    gp_Quaternion,
)
from OCP.GProp import GProp_GProps
from OCP.Standard import Standard_Failure, Standard_NoSuchObject
from OCP.TopLoc import TopLoc_Location
from OCP.TopoDS import TopoDS, TopoDS_Vertex, TopoDS_Iterator
from OCP.GeomAPI import GeomAPI_ProjectPointOnSurf
from ocp_tessellate.ocp_utils import get_faces, get_edges, get_vertices


downcast_LUT = {
    ta.TopAbs_VERTEX: TopoDS.Vertex_s,
    ta.TopAbs_EDGE: TopoDS.Edge_s,
    ta.TopAbs_WIRE: TopoDS.Wire_s,
    ta.TopAbs_FACE: TopoDS.Face_s,
    ta.TopAbs_SHELL: TopoDS.Shell_s,
    ta.TopAbs_SOLID: TopoDS.Solid_s,
    ta.TopAbs_COMPOUND: TopoDS.Compound_s,
    ta.TopAbs_COMPSOLID: TopoDS.CompSolid_s,
}

geom_LUT = {
    ta.TopAbs_VERTEX: "Vertex",
    ta.TopAbs_EDGE: BRepAdaptor_Curve,
    ta.TopAbs_WIRE: "Wire",
    ta.TopAbs_FACE: BRepAdaptor_Surface,
    ta.TopAbs_SHELL: "Shell",
    ta.TopAbs_SOLID: "Solid",
    ta.TopAbs_COMPOUND: "Compound",
    ta.TopAbs_COMPSOLID: "Compound",
}


geom_LUT_FACE = {
    ga.GeomAbs_Plane: "PLANE",
    ga.GeomAbs_Cylinder: "CYLINDER",
    ga.GeomAbs_Cone: "CONE",
    ga.GeomAbs_Sphere: "SPHERE",
    ga.GeomAbs_Torus: "TORUS",
    ga.GeomAbs_BezierSurface: "BEZIER",
    ga.GeomAbs_BSplineSurface: "BSPLINE",
    ga.GeomAbs_SurfaceOfRevolution: "REVOLUTION",
    ga.GeomAbs_SurfaceOfExtrusion: "EXTRUSION",
    ga.GeomAbs_OffsetSurface: "OFFSET",
    ga.GeomAbs_OtherSurface: "OTHER",
}

geom_LUT_EDGE = {
    ga.GeomAbs_Line: "LINE",
    ga.GeomAbs_Circle: "CIRCLE",
    ga.GeomAbs_Ellipse: "ELLIPSE",
    ga.GeomAbs_Hyperbola: "HYPERBOLA",
    ga.GeomAbs_Parabola: "PARABOLA",
    ga.GeomAbs_BezierCurve: "BEZIER",
    ga.GeomAbs_BSplineCurve: "BSPLINE",
    ga.GeomAbs_OffsetCurve: "OFFSET",
    ga.GeomAbs_OtherCurve: "OTHER",
}

shape_properties_LUT = {
    ta.TopAbs_VERTEX: None,
    ta.TopAbs_EDGE: BRepGProp.LinearProperties_s,
    ta.TopAbs_WIRE: BRepGProp.LinearProperties_s,
    ta.TopAbs_FACE: BRepGProp.SurfaceProperties_s,
    ta.TopAbs_SHELL: BRepGProp.SurfaceProperties_s,
    ta.TopAbs_SOLID: BRepGProp.VolumeProperties_s,
    ta.TopAbs_COMPOUND: BRepGProp.VolumeProperties_s,
    ta.TopAbs_COMPSOLID: BRepGProp.VolumeProperties_s,
}


def shapetype(obj):
    """Return TopoDS_Shape's TopAbs_ShapeEnum"""
    if obj.IsNull():
        raise ValueError("Null TopoDS_Shape object")

    return obj.ShapeType()


def downcast(obj):
    """Downcasts a TopoDS object to suitable specialized type

    Args:
      obj: TopoDS_Shape:

    Returns:

    """

    f_downcast = downcast_LUT[shapetype(obj)]
    return_value = f_downcast(obj)

    return return_value


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Vector
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Vector:
    def __init__(self, x, y=None, z=None):
        if isinstance(x, (gp_Pnt, gp_Dir)):
            self.wrapped = gp_Vec(x.XYZ())
        elif isinstance(x, gp_Vec):
            self.wrapped = x
        elif isinstance(x, (list, tuple)):
            self.wrapped = gp_Vec(*x)
        elif isinstance(x, Vector):
            self.wrapped = gp_Vec(x.wrapped.XYZ())
        else:
            self.wrapped = gp_Vec(x, y, z)

        self.X = self.wrapped.X()  # pylint: disable=invalid-name
        self.Y = self.wrapped.Y()  # pylint: disable=invalid-name
        self.Z = self.wrapped.Z()  # pylint: disable=invalid-name

    def normalized(self):
        return Vector(self.wrapped.Normalized())

    def center(self):
        return self

    @property
    def length(self):
        return self.wrapped.Magnitude()

    def sub(self, vec):
        if isinstance(vec, Vector):
            result = Vector(self.wrapped.Subtracted(vec.wrapped))
        elif isinstance(vec, tuple):
            result = Vector(self.wrapped.Subtracted(Vector(vec).wrapped))
        else:
            raise ValueError("Only Vectors or tuples can be subtracted from Vectors")

        return result

    def __sub__(self, vec):
        return self.sub(vec)

    def multiply(self, scale):
        return Vector(self.wrapped.Multiplied(scale))

    def __mul__(self, scale):
        return self.multiply(scale)

    def __truediv__(self, denom):
        return self.multiply(1.0 / denom)

    def __rmul__(self, scale):
        return self.multiply(scale)

    def to_tuple(self):
        return (self.X, self.Y, self.Z)

    def to_pnt(self):
        return gp_Pnt(self.wrapped.XYZ())

    def to_dir(self):
        return gp_Dir(self.wrapped.XYZ())

    def get_angle(self, vec):
        return math.degrees(self.wrapped.Angle(vec.wrapped))

    def __repr__(self):
        return "Vector: " + str((self.X, self.Y, self.Z))


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Axis(metaclass=AxisMeta)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class AxisMeta(type):
    @property
    def X(cls):  # pylint: disable=invalid-name
        return Axis((0, 0, 0), (1, 0, 0))

    @property
    def Y(cls):  # pylint: disable=invalid-name
        return Axis((0, 0, 0), (0, 1, 0))

    @property
    def Z(cls):  # pylint: disable=invalid-name
        return Axis((0, 0, 0), (0, 0, 1))


class Axis(metaclass=AxisMeta):
    def __init__(self, position=None, direction=None):
        self.position = Vector(0, 0, 0) if position is None else Vector(position)
        self.direction = Vector(0, 0, 1) if direction is None else Vector(direction)

    def __repr__(self):
        return f"({self.position.to_tuple()},{self.direction.to_tuple()})"

    @property
    def location(self):
        return Location(Plane(self.position, z_dir=self.direction))


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Matrix
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Matrix:
    def __init__(self):
        self.wrapped = gp_GTrsf()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Location
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Location:

    @property
    def position(self):
        return Vector(self.to_tuple()[0])

    def __init__(self, *args):
        # pylint: disable=too-many-branches
        transform = gp_Trsf()

        if len(args) == 0:
            pass

        elif len(args) == 1:
            translation = args[0]

            if isinstance(translation, (Vector, Iterable)):
                transform.SetTranslationPart(Vector(translation).wrapped)
            elif isinstance(translation, Plane):
                coordinate_system = gp_Ax3(
                    translation.origin.to_pnt(),
                    translation.z_dir.to_dir(),
                    translation.x_dir.to_dir(),
                )
                transform.SetTransformation(coordinate_system)
                transform.Invert()
            elif isinstance(args[0], Location):
                self.wrapped = translation.wrapped
                return
            elif isinstance(translation, TopLoc_Location):
                self.wrapped = translation
                return
            elif isinstance(translation, gp_Trsf):
                transform = translation
            else:
                raise TypeError("Unexpected parameters")

        elif len(args) == 2:
            if isinstance(args[0], (Vector, Iterable)):
                if isinstance(args[1], (Vector, Iterable)):
                    rotation = [math.radians(a) for a in args[1]]
                    quaternion = gp_Quaternion()
                    quaternion.SetEulerAngles(
                        gp_EulerSequence.gp_Intrinsic_XYZ, *rotation
                    )
                    transform.SetRotation(quaternion)
                elif isinstance(args[0], (Vector, tuple)) and isinstance(
                    args[1], (int, float)
                ):
                    angle = math.radians(args[1])
                    quaternion = gp_Quaternion()
                    quaternion.SetEulerAngles(
                        gp_EulerSequence.gp_Intrinsic_XYZ, 0, 0, angle
                    )
                    transform.SetRotation(quaternion)

                # set translation part after setting rotation (if exists)
                transform.SetTranslationPart(Vector(args[0]).wrapped)
            else:
                translation, origin = args
                coordinate_system = gp_Ax3(
                    Vector(origin).to_pnt(),
                    translation.z_dir.to_dir(),
                    translation.x_dir.to_dir(),
                )
                transform.SetTransformation(coordinate_system)
                transform.Invert()
        elif len(args) == 3:
            if (
                isinstance(args[0], (Vector, Iterable))
                and isinstance(args[1], (Vector, Iterable))
                and isinstance(args[2], (int, float))
            ):
                translation, axis, angle = args
                transform.SetRotation(
                    gp_Ax1(Vector().to_pnt(), Vector(axis).to_dir()),
                    math.radians(angle),
                )
            else:
                raise TypeError("Unsupported argument types for Location")

            transform.SetTranslationPart(Vector(translation).wrapped)
        self.wrapped = TopLoc_Location(transform)

    def inverse(self):
        return Location(self.wrapped.Inverted())

    def __mul__(self, other):
        if hasattr(other, "wrapped") and not isinstance(
            other.wrapped, TopLoc_Location
        ):  # Shape
            result = other.moved(self)
        elif isinstance(other, Iterable) and all(
            isinstance(o, Location) for o in other
        ):
            result = [Location(self.wrapped * loc.wrapped) for loc in other]
        else:
            result = Location(self.wrapped * other.wrapped)
        return result

    def to_tuple(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        transformation = self.wrapped.Transformation()
        trans = transformation.TranslationPart()
        rot = transformation.GetRotation()

        rv_trans = (trans.X(), trans.Y(), trans.Z())
        rv_rot = [
            math.degrees(a)
            for a in rot.GetEulerAngles(gp_EulerSequence.gp_Intrinsic_XYZ)
        ]

        return rv_trans, tuple(rv_rot)

    def __repr__(self):
        position_str = ", ".join((f"{v:.2f}" for v in self.to_tuple()[0]))
        orientation_str = ", ".join((f"{v:.2f}" for v in self.to_tuple()[1]))
        return f"(p=({position_str}), o=({orientation_str}))"

    def __str__(self):
        position_str = ", ".join((f"{v:.2f}" for v in self.to_tuple()[0]))
        orientation_str = ", ".join((f"{v:.2f}" for v in self.to_tuple()[1]))
        return f"Location: (position=({position_str}), orientation=({orientation_str}))"


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Plane
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Plane:
    def __init__(self, obj, x_dir=None, z_dir=None):
        if isinstance(obj, Vector):
            self.origin = obj
            self.z_dir = Vector(z_dir)
            if x_dir:
                self.x_dir = Vector(x_dir)
            else:
                ax3 = gp_Ax3(self.origin.to_pnt(), self.z_dir.to_dir())
                self.x_dir = Vector(ax3.XDirection()).normalized()

        elif isinstance(obj, Face):
            surface = BRep_Tool.Surface_s(obj.wrapped)
            if not isinstance(surface, Geom_Plane):
                raise ValueError("Planes can only be created from planar faces")
            properties = GProp_GProps()
            BRepGProp.SurfaceProperties_s(obj.wrapped, properties)
            self.origin = Vector(properties.CentreOfMass())
            self.x_dir = (
                Vector(x_dir)
                if x_dir
                else Vector(BRep_Tool.Surface_s(obj.wrapped).Position().XDirection())
            )
            self.z_dir = Plane.get_topods_face_normal(obj.wrapped)
        else:
            raise ValueError(f"Unable to create Plane from {type(obj)}")

    @property
    def location(self):
        return Location(self)

    @staticmethod
    def get_topods_face_normal(face):
        gp_pnt = gp_Pnt()
        normal = gp_Vec()
        projector = GeomAPI_ProjectPointOnSurf(gp_pnt, BRep_Tool.Surface_s(face))
        u_val, v_val = projector.LowerDistanceParameters()
        BRepGProp_Face(face).Normal(u_val, v_val, gp_pnt, normal)
        return Vector(normal)

    def to_local_coords(self, obj):
        # simplified version for Face only, not part of the original build123d
        l = Plane(Face(obj.wrapped)).location
        return Face(obj.wrapped.Moved(l.wrapped.Inverted()))


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# ShapeList(list)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class ShapeList(list):
    def filter_by(self, geom_type):
        return [obj for obj in self if obj.geom_type == geom_type]

    def sort_by(self, axis, reverse=False):
        axis_as_location = axis.location.inverse()
        objects = sorted(
            self,
            key=lambda o: (axis_as_location * Location(o.center())).position.Z,
            reverse=reverse,
        )
        return ShapeList(objects)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Shape
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Shape:
    def __init__(self, obj):
        self.wrapped = downcast(obj)

    @property
    def geom_type(self):
        shape = shapetype(self.wrapped)

        if shape == ta.TopAbs_EDGE:
            geom = geom_LUT_EDGE[BRepAdaptor_Curve(self.wrapped).GetType()]
        elif shape == ta.TopAbs_FACE:
            geom = geom_LUT_FACE[BRepAdaptor_Surface(self.wrapped).GetType()]
        elif shape == ta.TopAbs_VERTEX:
            geom = "VERTEX"
        elif shape == ta.TopAbs_SOLID:
            geom = "SOLID"
        else:
            geom = "UNKNOWN"

        return geom

    @property
    def area(self):
        properties = GProp_GProps()
        BRepGProp.SurfaceProperties_s(self.wrapped, properties)

        return properties.Mass()

    @staticmethod
    def compute_mass(obj):
        properties = GProp_GProps()
        calc_function = shape_properties_LUT[shapetype(obj.wrapped)]
        calc_function(obj.wrapped, properties)
        return properties.Mass()

    @property
    def volume(self):
        return Shape.compute_mass(self)

    @classmethod
    def cast(cls, obj):
        new_shape = None

        constructor__lut = {
            ta.TopAbs_VERTEX: Vertex,
            ta.TopAbs_EDGE: Edge,
            # ta.TopAbs_WIRE: Wire,
            ta.TopAbs_FACE: Face,
            ta.TopAbs_SHELL: Shell,
            ta.TopAbs_SOLID: Solid,
            ta.TopAbs_COMPOUND: Compound,
        }

        shape_type = shapetype(obj)
        new_shape = constructor__lut[shape_type](downcast(obj))

        return new_shape

    def transform_shape(self, t_matrix):
        if isinstance(self, Vertex):
            new_shape = Vertex(*t_matrix.multiply(Vector(self)))
        else:
            transformed = Shape.cast(
                BRepBuilderAPI_Transform(self.wrapped, t_matrix.wrapped.Trsf()).Shape()
            )
            new_shape = copy.deepcopy(self, None)
            new_shape.wrapped = transformed.wrapped

        return new_shape

    def edges(self):
        return ShapeList(
            [
                Edge(obj)
                for obj in ShapeList(get_edges(self.wrapped))
                if not BRep_Tool.Degenerated_s(TopoDS.Edge_s(obj))
            ]
        )

    def vertices(self):
        return ShapeList([Vertex(obj) for obj in ShapeList(get_vertices(self.wrapped))])

    def faces(self):
        return ShapeList([Face(obj) for obj in ShapeList(get_faces(self.wrapped))])


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Compound(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


def unwrapped_shapetype(obj: Shape):
    """Return Shape's TopAbs_ShapeEnum"""
    if isinstance(obj, Compound):
        shapetypes = set(shapetype(o.wrapped) for o in obj)
        if len(shapetypes) == 1:
            result = shapetypes.pop()
        else:
            result = shapetype(obj)
    else:
        result = shapetype(obj.wrapped)
    return result


class Compound(Shape):
    def center(self):
        properties = GProp_GProps()
        calc_function = shape_properties_LUT[unwrapped_shapetype(self)]
        if calc_function:
            calc_function(self.wrapped, properties)
            middle = Vector(properties.CentreOfMass())
        return middle

    def __iter__(self):
        iterator = TopoDS_Iterator(self.wrapped)

        while iterator.More():
            yield Shape.cast(iterator.Value())
            iterator.Next()


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Solid(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Solid(Shape):
    def center(self):
        properties = GProp_GProps()
        calc_function = shape_properties_LUT[shapetype(self.wrapped)]
        if calc_function:
            calc_function(self.wrapped, properties)
            middle = Vector(properties.CentreOfMass())
        else:
            raise NotImplementedError
        return middle


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Shell(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Shell(Shape):
    def center(self):
        properties = GProp_GProps()
        BRepGProp.LinearProperties_s(self.wrapped, properties)
        return Vector(properties.CentreOfMass())


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Face(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Face(Shape):
    def _geom_adaptor(self):
        return BRep_Tool.Surface_s(self.wrapped)

    @property
    def length(self):
        result = None
        if self.geom_type == "PLANE":
            flat_face = Plane(self).to_local_coords(self)
            face_vertices = flat_face.vertices().sort_by(Axis.X)
            result = face_vertices[-1].X - face_vertices[0].X
        return result

    @property
    def width(self):
        result = None
        if self.geom_type == "PLANE":
            flat_face = Plane(self).to_local_coords(self)
            face_vertices = flat_face.vertices().sort_by(Axis.Y)
            result = face_vertices[-1].Y - face_vertices[0].Y
        return result

    def _uv_bounds(self):
        return BRepTools.UVBounds_s(self.wrapped)

    def center(self):
        if self.geom_type == "PLANE":
            properties = GProp_GProps()
            BRepGProp.SurfaceProperties_s(self.wrapped, properties)
            center_point = properties.CentreOfMass()

        else:
            u_val0, u_val1, v_val0, v_val1 = self._uv_bounds()
            u_val = 0.5 * (u_val0 + u_val1)
            v_val = 0.5 * (v_val0 + v_val1)

            center_point = gp_Pnt()
            normal = gp_Vec()
            BRepGProp_Face(self.wrapped).Normal(u_val, v_val, center_point, normal)

        return Vector(center_point)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Edge(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Edge(Shape):
    def _geom_adaptor(self):
        return BRepAdaptor_Curve(self.wrapped)

    @property
    def radius(self):
        geom = self._geom_adaptor()
        try:
            circ = geom.Circle()
        except (Standard_NoSuchObject, Standard_Failure) as err:
            raise ValueError("Shape could not be reduced to a circle") from err
        return circ.Radius()

    @property
    def length(self):
        return GCPnts_AbscissaPoint.Length_s(self._geom_adaptor())

    def tangent_at(
        self,
        location_param=0.5,
    ):
        curve = self._geom_adaptor()

        tmp = gp_Pnt()
        res = gp_Vec()

        curve.D1(location_param, tmp, res)

        return Vector(res)

    def normal(self):
        curve = self._geom_adaptor()
        gtype = self.geom_type

        if gtype == "CIRCLE":
            circ = curve.Circle()
            return_value = Vector(circ.Axis().Direction())
        elif gtype == "ELLIPSE":
            ell = curve.Ellipse()
            return_value = Vector(ell.Axis().Direction())
        else:
            find_surface = BRepLib_FindSurface(self.wrapped, OnlyPlane=True)
            surf = find_surface.Surface()

            if isinstance(surf, Geom_Plane):
                pln = surf.Pln()
                return_value = Vector(pln.Axis().Direction())
            else:
                raise ValueError("Normal not defined")

        return return_value

    def position_at(self, distance: float) -> Vector:
        curve = self._geom_adaptor()
        param = self.param_at(distance)
        return Vector(curve.Value(param))

    def param_at(self, distance: float) -> float:
        curve = self._geom_adaptor()

        length = GCPnts_AbscissaPoint.Length_s(curve)
        return GCPnts_AbscissaPoint(
            curve, length * distance, curve.FirstParameter()
        ).Parameter()

    def __mod__(self, position):
        return self.tangent_at(position)

    def __matmul__(self, position):
        return self.position_at(position)

    def center(self):
        return self.position_at(0.5)

    @property
    def arc_center(self):
        geom_type = self.geom_type
        geom_adaptor = self._geom_adaptor()

        if geom_type == "CIRCLE":
            return_value = Vector(geom_adaptor.Circle().Position().Location())
        elif geom_type == "ELLIPSE":
            return_value = Vector(geom_adaptor.Ellipse().Position().Location())
        else:
            raise ValueError(f"{geom_type} has no arc center")

        return return_value


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Vertex(Shape)
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class Vertex(Shape):
    def __init__(self, x, y=None, z=None, ocp_vx=None):
        if isinstance(x, (tuple, list)):
            x, y, z = x
        elif isinstance(self, Vector):
            x, y, z = x.to_tuple()
        elif isinstance(x, Vertex):
            x, y, z = x.X, x.Y, x.Z
        elif isinstance(x, TopoDS_Vertex):
            ocp_vx = x

        ocp_vx = (
            downcast(BRepBuilderAPI_MakeVertex(gp_Pnt(x, y, z)).Vertex())
            if ocp_vx is None
            else ocp_vx
        )

        super().__init__(ocp_vx)
        self.X, self.Y, self.Z = self.to_tuple()  # pylint: disable=invalid-name

    def to_tuple(self):
        geom_point = BRep_Tool.Pnt_s(self.wrapped)
        return (geom_point.X(), geom_point.Y(), geom_point.Z())

    def center(self):
        return Vector(*self.to_tuple())



"""
state uses "poor man's locking": nodejs does not support fcntl (POSIX) or 
LockFileEx (Windows) locking. This module implements a simple locking 
mechanism for nodejs and Python leveraging "mkdir" to create locks. 

Ideas are taken from the node js library "proper-lockfile"
(https://github.com/moxystudio/node-proper-lockfile)
"""
import json
import os
import time

STALE_DURATION_MS = 3000  # 3 seconds
RETRIES = 7
INTERVAL_MS = 500  # INTERVAL_MS * RETRIES > STALE_DURATION
CONFIG_FILE = "~/.ocpvscode"


def get_config_file():
    """Get the config file"""
    return resolve_path(CONFIG_FILE)


def get_lock_file(file):
    """Get the lock file for the given file"""
    return f"{file}.lock"


def resolve_path(file):
    """Resolve the canonical path of the file"""
    return os.path.realpath(os.path.expanduser(file))


def is_lock_stale(mtime, stale_seconds):
    """Check if the lock is stale"""
    return (time.time() - mtime) > stale_seconds


def remove_lock(lockfile):
    """Remove the lockfile"""
    try:
        os.rmdir(lockfile)
    except FileNotFoundError:
        print(f"Lock file {lockfile} not found")
    except Exception as ex:
        raise RuntimeError(f"Unable to remove lock file {lockfile}") from ex


def acquire_lock(
    lockfile,
    retries=RETRIES,
    interval_ms=INTERVAL_MS,
    stale_duration_ms=STALE_DURATION_MS,
    retry=0,
):
    """Acquire the lock for the given file"""

    # Use mkdir to create the lockfile
    try:
        os.mkdir(lockfile)
        # lock successfully acquired
        # print(f"Lock file {lockfile} aquired")
    except FileExistsError:
        print(f"Lock file {lockfile} already exists")
        if is_lock_stale(os.stat(lockfile).st_mtime, stale_duration_ms / 1000):
            print(f"Lock file {lockfile} is stale")
            try:
                # assume the lockfile is stale and simply remove it
                remove_lock(lockfile)
            except FileNotFoundError:
                pass  # lock seems to be just removed, which is ok
            except Exception as ex:
                raise RuntimeError(
                    f"Unable to remove stale lock file {lockfile}."
                ) from ex

            # try to acquire the lock again
            acquire_lock(lockfile, retries, interval_ms, stale_duration_ms, retry + 1)
        else:
            if retry < retries:
                time.sleep(interval_ms / 1000)
                print("Retrying to acquire lock")
                acquire_lock(
                    lockfile, retries, interval_ms, stale_duration_ms, retry + 1
                )
            else:
                raise RuntimeError(  # pylint: disable=raise-missing-from
                    f"Unable to acquire lock for file {lockfile} after {retries} retries"
                )

    except Exception as ex:
        raise RuntimeError(f"Cannot create lock file {lockfile}.") from ex


def lock(file, retries=RETRIES, interval_ms=INTERVAL_MS):
    """Lock the given file"""
    lockfile = get_lock_file(file)
    acquire_lock(lockfile, retries, interval_ms)


def unlock(file):
    """Unlock the given file"""
    lockfile = get_lock_file(file)
    remove_lock(lockfile)


def update_state(port, key=None, value=None):
    """Update the config file with the given key/value pair"""

    config_file = resolve_path(CONFIG_FILE)
    lock(config_file)
    try:
        # pylint: disable=consider-using-with
        fd = open(config_file, "r+", encoding="utf-8")
        config = fd.read()
        if config == "":
            config = {}
        else:
            config = json.loads(config)
    except FileNotFoundError:
        # pylint: disable=consider-using-with
        fd = open(config_file, "w+", encoding="utf-8")
        config = {}
    except Exception as ex:
        unlock(config_file)
        raise RuntimeError(f"Unable to open config file {config_file}.") from ex
    fd.close()

    port = str(port)
    if config.get(port) is None:
        config[port] = {}

    if key is None:
        del config[port]
    elif value is None:
        del config[port][key]
    elif isinstance(value, str):
        config[port][key] = value.rstrip(os.path.sep)
    else:
        config[port][key] = [v.rstrip(os.path.sep) for v in value]

    try:
        with open(config_file, "w", encoding="utf-8") as fd:
            fd.write(json.dumps(config, indent=2))
    except Exception as ex:
        raise RuntimeError(f"Unable to write config file {config}.") from ex
    finally:
        unlock(config_file)


def get_state():
    """Get the config file"""
    config_file = resolve_path(CONFIG_FILE)
    lock(config_file)
    try:
        with open(config_file, "r", encoding="utf-8") as fd:
            config = fd.read()
            if config == "":
                config = {}
            else:
                config = json.loads(config)
        unlock(config_file)
    except FileNotFoundError as ex:
        raise RuntimeError(f"Unable to open config file {config_file}.") from ex

    return config
